package fr.insa.clavardator.ui;

import com.jfoenix.controls.JFXSnackbar;
import fr.insa.clavardator.ui.chat.ChatController;
import fr.insa.clavardator.ui.dialogs.AboutDialogController;
import fr.insa.clavardator.ui.dialogs.EditUsernameDialogController;
import fr.insa.clavardator.ui.dialogs.SnackbarController;
import fr.insa.clavardator.ui.users.UserListController;
import fr.insa.clavardator.users.CurrentUser;
import fr.insa.clavardator.users.UserList;
import fr.insa.clavardator.util.Log;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.layout.StackPane;

import java.io.IOException;
import java.net.SocketException;
import java.net.URL;
import java.util.ResourceBundle;

public class MainController implements Initializable {

	private final CurrentUser currentUser;
	@FXML
	private StackPane root;
	@FXML
	private StackPane mainContainer;
	@FXML
	private ToolbarController toolbarController;
	@FXML
	private EditUsernameDialogController editUserDialogController;
	@FXML
	private AboutDialogController aboutDialogController;
	@FXML
	private UserListController listController;
	@FXML
	private ChatController chatController;
	@FXML
	private LoadingScreenController loadingController;
	@FXML
	private ErrorScreenController errorController;

	private JFXSnackbar snackbar;
	private UserList userList;

	public MainController() {
		currentUser = CurrentUser.getInstance();
		currentUser.addObserver(propertyChangeEvent -> {
			if (propertyChangeEvent.getPropertyName().equals("state")) {
				final CurrentUser.State newState = (CurrentUser.State) propertyChangeEvent.getNewValue();
				onCurrentUserStateChange(newState);
			}
		});
	}

	private void onCurrentUserStateChange(CurrentUser.State newState) {
		if (newState == CurrentUser.State.VALID)
			startChat();
		else
			Platform.runLater(() -> showLogin(newState == CurrentUser.State.INVALID));
	}

	private void openEditUsernameDialog(EditUsernameDialogController.Mode mode) {
		final boolean initialMode = mode == EditUsernameDialogController.Mode.INITIAL;
		editUserDialogController.setOnCancelListener(() -> {
			if (initialMode) {
				System.exit(0);
			} else {
				showChat();
			}
		});
		editUserDialogController.setOnSuccessListener(() -> {
			if (!initialMode) {
				showSnackbarEvent("Nom d'utilisateur changé !", SnackbarController.Mode.SUCCESS);
			}
		});
		editUserDialogController.show(root, mode);
	}

	private void showSnackbarEvent(String text, SnackbarController.Mode mode) {
		try {
			final FXMLLoader loader = new FXMLLoader(getClass().getResource("dialogs/snackbar.fxml"));
			snackbar.enqueue(new JFXSnackbar.SnackbarEvent(
					loader.load()
			));
			final SnackbarController controller = loader.getController();
			controller.setText(text);
			controller.setMode(mode);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void openAboutDialog() {
		aboutDialogController.show(root);
	}

	private void discoverActiveUsers() {
		if (userList != null) {
			userList.discoverActiveUsers((e) -> CurrentUser.getInstance().setState(CurrentUser.State.INVALID));
		}
	}

	private void startListening() {
		if (userList != null) {
			userList.startDiscoveryListening();
			userList.startUserListening((e) ->
					Log.e(this.getClass().getSimpleName(), "Error listening to users", e));
		}
	}

	private void startChat() {
		Log.v(this.getClass().getSimpleName(), "Chat started");
		discoverActiveUsers();
		startListening();
		Platform.runLater(this::showChat);
	}

	private void showChat() {
		Log.v(this.getClass().getSimpleName(), "Chat shown");
		loadingController.hide();
		mainContainer.setVisible(true);
	}

	private void showLogin(boolean isError) {
		Log.v(this.getClass().getSimpleName(), "Login shown");
		mainContainer.setVisible(false);
		loadingController.hide();
		openEditUsernameDialog(isError ? EditUsernameDialogController.Mode.ERROR : EditUsernameDialogController.Mode.INITIAL);
	}

	private void showError() {
		Log.v(this.getClass().getSimpleName(), "Error shown");
		mainContainer.setVisible(false);
		loadingController.hide();
		errorController.show();
	}

	@Override
	public void initialize(URL url, ResourceBundle rb) {
		loadingController.show("Initialisation de Clavardator...");
		snackbar = new JFXSnackbar(root);

		try {
			currentUser.init();
			Log.v("INIT", "Current user: " + currentUser.getId());
		} catch (SocketException e) {
			Log.e("INIT", "Could not init user", e);
			showError();
		}

		listController.setUserSelectedListener((user) -> chatController.setRemoteUser(user));
		chatController.addAttachmentListener(() -> System.out.println("attach event"));
		chatController.addSendErrorListener((e) -> showSnackbarEvent("Erreur: Message non envoyé", SnackbarController.Mode.ERROR));
		toolbarController.addEditListener(() -> openEditUsernameDialog(EditUsernameDialogController.Mode.EDIT));
		toolbarController.addAboutListener(this::openAboutDialog);
	}

	public void setUserList(UserList userList) {
		this.userList = userList;
		listController.setUserList(userList);
		listController.setRefreshUserListener(this::discoverActiveUsers);
	}
}
