package fr.insa.clavardator.users;

import fr.insa.clavardator.chat.ChatHistory;
import fr.insa.clavardator.chat.Message;
import fr.insa.clavardator.network.PeerConnection;
import fr.insa.clavardator.util.ErrorCallback;
import fr.insa.clavardator.util.Log;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.EOFException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Date;

public class PeerUser extends User implements Comparable<PeerUser> {
	protected transient ChatHistory history;
	private State state = State.DISCONNECTED;
	private transient PeerConnection connection;

	public PeerUser(int id, String username) {
		super(id, username);
		history = new ChatHistory(this);
	}

	public PeerUser(int id) {
		super(id);
		history = new ChatHistory(this);
	}

	/**
	 * Asynchronously connects to the user and receives its information (id, username)
	 *
	 * @param ipAddr        The IP address of this user
	 * @param callback      The function to call on success
	 * @param errorCallback The function to call on socket error
	 */
	public void connect(InetAddress ipAddr, UserConnectedCallback callback, ErrorCallback errorCallback) {
		if (connection != null && connection.isOpen()) {
			connection.close();
		}
		Log.v(this.getClass().getSimpleName(), "Creating new TCP connection with " + id);
		// Connect to the peer
		setState(State.CONNECTING);
		connection = new PeerConnection(
				ipAddr,
				(thisConnection) -> init(thisConnection, callback, errorCallback),
				e -> {
					Log.e(this.getClass().getSimpleName(), "Could not create TCP connection with " + id, e);
					disconnect();
					errorCallback.onError(e);
				});
	}

	/**
	 * Asynchronously connects to the user using the socket and receives its information (id, username)
	 *
	 * @param socket        A socket already connected to the user
	 * @param callback      The function to call on success, with the new ActiveUser as parameter
	 * @param errorCallback The function to call on socket error
	 */
	public void connect(Socket socket, UserConnectedCallback callback, ErrorCallback errorCallback) {
		if (connection != null && connection.isOpen()) {
			connection.close();
		}
		setState(State.CONNECTING);
		connection = new PeerConnection(socket);

		init(connection, callback, errorCallback);
	}

	public void sendTextMessage(String msg,  @Nullable ErrorCallback errorCallback) {
		if (connection != null) {
			Log.v(this.getClass().getSimpleName(),
					"Sending message to " + this.getUsername() + " / " + this.getId() + ": " + msg);
			final Message message = new Message(CurrentUser.getInstance(), this, new Date(), msg);
			connection.send(
					message,
					() -> history.addMessage(message),
					errorCallback);
		} else {
			Log.e(this.getClass().getSimpleName(), "Could not send message: connection is not initialized");
		}
	}

	private void init(PeerConnection thisConnection, UserConnectedCallback callback, ErrorCallback errorCallback) {
		// Send our username
		String currentUserUsername = CurrentUser.getInstance().getUsername();
		int currentUserId = CurrentUser.getInstance().getId();
		thisConnection.send(new UserInformation(currentUserId, currentUserUsername), null, e -> {
			disconnect();
			errorCallback.onError(e);
		});

		// Receive peer's username
		thisConnection.receiveOne(msg -> {
			assert msg instanceof UserInformation;
			UserInformation userInfo = (UserInformation) msg;
			// TODO : Check username unique
			assert id == userInfo.id;
			setUsername(userInfo.getUsername());
			callback.onUserConnected();

			// Subscribe to incoming messages
			subscribeToMessages(e -> {
				disconnect();
				errorCallback.onError(e);
			});
			setState(State.CONNECTED);

		}, e -> {
			disconnect();
			errorCallback.onError(e);
		});
	}


	private void subscribeToMessages(ErrorCallback errorCallback) {
		connection.receive(
				msg -> {
					Log.v(this.getClass().getSimpleName(), "Received message from " + id);
					if (msg instanceof UserInformation) {
						assert ((UserInformation) msg).id == getId();
						Log.v(this.getClass().getSimpleName(), "Message username: " + ((UserInformation) msg).getUsername());
						setUsername(((UserInformation) msg).getUsername());
					} else if (msg instanceof Message) {
						assert ((Message) msg).getRecipient().id != id;
						Log.v(this.getClass().getSimpleName(), "Message text: " + ((Message) msg).getText());
						history.addMessage((Message) msg);
					}
				},
				e -> {
					if (e instanceof EOFException) {
						disconnect();
					} else {
						Log.e(this.getClass().getSimpleName(), "Error receiving message from " + id, e);
						errorCallback.onError(e);
					}
				});
	}


	public void disconnect() {
		Log.v(this.getClass().getSimpleName(), "Disconnecting from user: " + id);
		if (connection != null && connection.isOpen()) {
			connection.close();
			connection = null;
		}
		setState(State.DISCONNECTED);
	}


	/**
	 * Get the value of history
	 *
	 * @return the value of history
	 */
	public ChatHistory getHistory() {
		return history;
	}

	private void setState(State state) {
		pcs.firePropertyChange("state", this.state, state);
		this.state = state;
	}

	public boolean isActive() {
		return state == State.CONNECTED;
	}

	@Override
	public int compareTo(@NotNull PeerUser peerUser) {
		if (peerUser.isActive() && !this.isActive()) {
			return 1;
		} else if (!peerUser.isActive() && this.isActive()) {
			return -1;
		}
		return getUsername().compareTo(peerUser.getUsername());
	}

	public enum State {
		CONNECTING,
		CONNECTED,
		DISCONNECTED,
	}

	public interface UserConnectedCallback {
		void onUserConnected();
	}
}
