package fr.insa.clavardator.users;

import fr.insa.clavardator.db.DatabaseController;
import fr.insa.clavardator.network.ConnectionListener;
import fr.insa.clavardator.network.NetDiscoverer;
import fr.insa.clavardator.util.ErrorCallback;
import fr.insa.clavardator.util.Log;

import java.beans.PropertyChangeEvent;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Predicate;

import static fr.insa.clavardator.network.NetUtil.getIdFromIp;

public class UserList {

	private final Map<Integer, PeerUser> inactiveUsers = new HashMap<>();
	private final Map<Integer, PeerUser> activeUsers = new HashMap<>();

	private final ArrayList<UserConnectionCallback> newUsersObservers = new ArrayList<>();

	private final DatabaseController db = new DatabaseController();
	private final NetDiscoverer netDiscoverer = new NetDiscoverer();
	private final ConnectionListener connectionListener = new ConnectionListener();

	public UserList() {
	}

	/**
	 * Discover all active users over the network. Observers are notified for each user discovered.
	 *
	 * @param errorCallback The function to call on error
	 * @see UserList#addNewUserObserver(UserConnectionCallback)
	 */
	public void discoverActiveUsers(ErrorCallback errorCallback) {
		netDiscoverer.discoverActiveUsers("CLAVARDATOR_BROADCAST", (ipAddr, data) -> {
			int id = getIdFromIp(ipAddr);
			Log.v(this.getClass().getSimpleName(), "Discovered new user: " + id);
			final PeerUser finalUser = createNewUser(id);
			if (finalUser != null) {
				finalUser.connect(ipAddr, () ->
						finalUser.addObserver(evt -> userChangeObserver(finalUser, evt)), errorCallback);
			}
		}, errorCallback);
	}

	public void startDiscoveryListening() {
		netDiscoverer.startDiscoveryListening(
				"CLAVARDATOR_RESPONSE",
				null,
				Throwable::printStackTrace);
	}

	public void startUserListening(ErrorCallback errorCallback) {
		connectionListener.acceptConnection(
				(clientSocket) -> {
					int id = getIdFromIp(clientSocket.getInetAddress());
					Log.v(this.getClass().getSimpleName(), "new connection from user: " + id);
					final PeerUser finalUser = createNewUser(id);
					if (finalUser != null) {
						finalUser.connect(clientSocket, () ->
								finalUser.addObserver(evt -> userChangeObserver(finalUser, evt)), errorCallback);
					}
				},
				errorCallback);
	}

	private PeerUser createNewUser(int id) {
		// If already connected, warn and return
		if (activeUsers.containsKey(id)) {
			Log.w(getClass().getSimpleName(),
					"An already connected user tried to initiate a new connection: user id " + id);
			return null;
		}

		// Get the user if already existing
		PeerUser user = inactiveUsers.get(id);
		// else create it
		if (user == null) {
			// Username is set on TCP connection start
			user = new PeerUser(id);
			inactiveUsers.put(id, user);
			notifyNewUserObservers(user);
		}

		return user;
	}

	private void userChangeObserver(PeerUser user, PropertyChangeEvent evt) {
		int id = user.id;

		if (evt.getPropertyName().equals("state")) {
			PeerUser.State oldState = (PeerUser.State) evt.getOldValue();
			PeerUser.State newState = (PeerUser.State) evt.getNewValue();

			if ((oldState == PeerUser.State.DISCONNECTED || oldState == PeerUser.State.CONNECTING) &&
					newState == PeerUser.State.CONNECTED) {
				// Connection handling
				if (!inactiveUsers.containsKey(id)) {
					if (activeUsers.containsKey(id)) {
						Log.w(getClass().getSimpleName(), "Tried to set state CONNECTED on an already connected user: user id " + id);
					} else {
						Log.w(getClass().getSimpleName(), "Tried to set state CONNECTED on an unknown user: user id " + id);
					}
					return;
				}

				inactiveUsers.remove(id);
				activeUsers.put(id, user);
				Log.v(getClass().getSimpleName(), "State of user " + id + " updated from " +
						oldState.toString() + " to " + newState.toString());
			} else if (oldState == PeerUser.State.CONNECTED &&
					(newState == PeerUser.State.DISCONNECTED || newState == PeerUser.State.CONNECTING)) {
				// Disconnection
				if (!activeUsers.containsKey(id)) {
					if (inactiveUsers.containsKey(id)) {
						Log.e(getClass().getSimpleName(), "Tried to set state DISCONNECTED on an already disconnected user: user id " + id);
					} else {
						Log.e(getClass().getSimpleName(), "Tried to set state DISCONNECTED on an unknown user: user id " + id);
					}
					return;
				}

				activeUsers.remove(id);
				inactiveUsers.put(id, user);
				Log.v(getClass().getSimpleName(), "State of user " + id + " updated from " +
						oldState.toString() + " to " + newState.toString());
			}
		}
	}

	public void addNewUserObserver(UserConnectionCallback connectionCallback) {
		newUsersObservers.add(connectionCallback);
	}

	private void notifyNewUserObservers(PeerUser user) {
		newUsersObservers.forEach(callback -> callback.onUserConnected(user));
	}

	/**
	 * Tests locally if a username is available
	 *
	 * @param username The username to test
	 * @return True if the username is available
	 */
	public boolean isUsernameAvailable(String username) {
		Predicate<User> usernameEqual = user -> user.getUsername().equals(username);
		return activeUsers.values().stream().noneMatch(usernameEqual) &&
				inactiveUsers.values().stream().noneMatch(usernameEqual);
	}

	/**
	 * Tell all active users that our username changed
	 *
	 * @param username The new username
	 */
	public void propagateUsernameChange(String username) {
		// TODO
	}

	/**
	 * Close all running threads, sockets and db connection
	 * Must be called before exiting the app
	 */
	public void destroy() {
		netDiscoverer.stopDiscovery();
		connectionListener.stopAccepting();
//		db.close();
		for (PeerUser user : activeUsers.values()) {
			user.disconnect();
		}
	}

	public interface UserConnectionCallback {
		void onUserConnected(PeerUser user);
	}

	public interface UserDisconnectionCallback {
		void onUserDisconnected(PeerUser user);
	}

}
